from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="fastlife_")

    fastlife_route_prefix: str = Field(default="/_fl")
    template_search_path: str = Field(default="fastlife:templates")
    registry_class: str = Field(default="fastlife.configurator.registry:AppRegistry")
    template_renderer_class: str = Field(
        default="fastlife.templating.renderer.jinja2:Jinja2TemplateRenderer"
    )
    form_data_model_prefix: str = Field(default="payload")
    csrf_token_name: str = Field(default="csrf_token")



from .binding import Template, template
from .renderer import AbstractTemplateRenderer, Jinja2TemplateRenderer

__all__ = [
    "AbstractTemplateRenderer",
    "Jinja2TemplateRenderer",
    "Template",
    "template",
]



from typing import Sequence

from markupsafe import Markup

from fastlife.templating.renderer.abstract import AbstractTemplateRenderer

from .base import Widget


class ModelWidget(Widget):
    def __init__(
        self,
        name: str,
        *,
        children_widget: Sequence[Widget],
        removable: bool,
        title: str,
        token: str,
    ):
        super().__init__(name, title=title, removable=removable, token=token)
        self.children_widget = children_widget

    def get_template(self) -> str:
        return "pydantic_form/model.jinja2"

    async def to_html(self, renderer: AbstractTemplateRenderer) -> Markup:
        """Return the html version"""
        children_widget = [
            await child.to_html(renderer) for child in self.children_widget
        ]
        return Markup(
            await renderer.render_template(
                self.get_template(), widget=self, children_widget=children_widget
            )
        )



from typing import Any, AsyncGenerator, Callable, Mapping, Optional, Sequence, Type

from fastapi import Request
from jinja2 import Environment, FileSystemLoader, Template
from markupsafe import Markup

from fastlife.configurator.settings import Settings
from fastlife.shared_utils.resolver import resolve_path

from .abstract import AbstractTemplateRenderer
from .widgets.factory import WidgetFactory


def build_searchpath(template_search_path: str) -> Sequence[str]:
    searchpath: list[str] = []
    paths = template_search_path.split(",")

    for path in paths:
        if ":" in path:
            searchpath.append(resolve_path(path))
        else:
            searchpath.append(path)
    return searchpath


class Jinja2TemplateRenderer(AbstractTemplateRenderer):
    def __init__(self, settings: Settings) -> None:
        super().__init__()
        self.route_prefix = settings.fastlife_route_prefix
        self.env = Environment(
            loader=FileSystemLoader(build_searchpath(settings.template_search_path)),
            enable_async=True,
        )
        self.form_data_model_prefix = settings.form_data_model_prefix
        self.csrf_token_name = settings.csrf_token_name

    def _get_template(self, template: str, **kwargs: Any) -> Template:
        return self.env.get_template(
            template,
            globals={**kwargs},
        )

    def get_csrf_token(self, request: Request) -> Callable[..., str]:
        def get_csrf_token() -> str:
            return request.scope.get(self.csrf_token_name, "")

        return get_csrf_token

    async def _render_block(self, gen_blocks: AsyncGenerator[str, None]) -> str:
        # the typing of Jinja2 async is wrong. It says there is Iterator[str]
        # insead of Asyngenerator[str, None]
        blocks: list[str] = []
        async for value in gen_blocks:
            blocks.append(value)
        return self.env.concat(blocks)  # type: ignore

    async def render_page(self, request: Request, template: str, **kwargs: Any) -> str:
        """
        Render the the template to build a full page or only a block, in case of
        htmx request containing a HX-Target.
        """
        tpl = self._get_template(
            template,
            request=request,
            get_csrf_token_name=lambda: self.csrf_token_name,
            get_csrf_token=self.get_csrf_token(request),
            pydantic_form=self.pydantic_form,
        )
        if "HX-Target" in request.headers:
            block_name = request.headers["HX-Target"]
            # We render the hx-target as a Jinja2 block of the page,
            # we use low lever functions here to build only whats we need.
            # Fist, we need to push the macros in the context has it is done in the
            # base.jinja2 for compatibility.
            macros = self._get_template("globals.jinja2", request=request)
            ctx = tpl.new_context(kwargs)
            render_macros = macros.root_render_func(ctx)
            # the typing of Jinja2 async is wrong here
            gen_blocks = await self._render_block(render_macros)  # type: ignore
            # Now we build the block without the layout.
            render_block = tpl.blocks[block_name]
            gen_blocks = render_block(ctx)  # type: ignore
            return await self._render_block(gen_blocks)  # type: ignore
        else:
            # we render the full page
            ret = await tpl.render_async(**kwargs)
        return ret

    async def render_template(self, template: str, **kwargs: Any) -> str:
        tpl = self._get_template(
            template,
            pydantic_form=self.pydantic_form,
        )
        ret = await tpl.render_async(**kwargs)
        return ret

    async def pydantic_form(
        self,
        model: Type[Any],
        form_data: Optional[Mapping[str, Any]] = None,
        name: Optional[str] = None,
        token: Optional[str] = None,
        removable: bool = False,
    ) -> Markup:
        return await WidgetFactory(self, token).get_markup(
            model,
            form_data or {},
            prefix=(name or self.form_data_model_prefix),
            removable=removable,
        )



from typing import Optional

from fastapi import Query, Response

from fastlife import Configurator, configure
from fastlife.configurator.registry import Registry
from fastlife.shared_utils.resolver import resolve_extended


async def show_widget(
    typ: str,
    reg: Registry,
    name: Optional[str] = Query(...),
    token: Optional[str] = Query(...),
    removable: bool = Query(False),
) -> Response:
    model_cls = resolve_extended(typ)
    data = await reg.renderer.pydantic_form(model_cls, None, name, token, removable)
    return Response(data, headers={"Content-Type": "text/html"})


@configure
def includeme(config: Configurator) -> None:
    route_prefix = config.registry.settings.fastlife_route_prefix
    config.add_route(
        f"{route_prefix}/pydantic-form/widgets/{{typ}}", show_widget, methods=["GET"]
    )



import re
from typing import Any, Mapping
from urllib.parse import urlencode

import bs4
import httpx
from fastapi.testclient import TestClient
from starlette.types import ASGIApp


class WebForm:
    def __init__(self, client: "WebTestClient", origin: str, form: bs4.Tag):
        self._client = client
        self._form = form
        self._origin = origin
        self._formdata: dict[str, str] = {}
        inputs = self._form.find_all("input")
        for input in inputs:
            if input.attrs.get("type") == "checkbox" and "checked" not in input.attrs:
                continue
            self._formdata[input.attrs["name"]] = input.attrs.get("value", "")
        # field select, textearea...

    def set(self, fieldname: str, value: str) -> Any:
        if fieldname not in self._formdata:
            raise ValueError(f"{fieldname} does not exists")
        self._formdata[fieldname] = value

    def button(self, text: str) -> "WebForm":
        assert self._form.find("button", string=re.compile(f".*{text}.*")) is not None
        return self

    def submit(self) -> "WebResponse":
        target = (
            self._form.attrs.get("hx-post")
            or self._form.attrs.get("post")
            or self._origin
        )
        return self._client.post(target, data=self._formdata)


class WebResponse:
    def __init__(self, client: "WebTestClient", origin: str, response: httpx.Response):
        self._client = client
        self._response = response
        self._origin = origin
        self._html: bs4.BeautifulSoup | None = None
        self._form: WebForm | None = None

    @property
    def status_code(self) -> int:
        return self._response.status_code

    @property
    def content_type(self) -> str:
        return self._response.headers["content-type"]

    @property
    def headers(self) -> httpx.Headers:
        return self._response.headers

    @property
    def text(self) -> str:
        return self._response.text

    @property
    def html(self) -> bs4.Tag:
        if self._html is None:
            self._html = bs4.BeautifulSoup(self._response.text, "html.parser")
        return self._html

    @property
    def html_body(self) -> bs4.Tag:
        body = self.html.find("body")
        assert body is not None
        assert isinstance(body, bs4.Tag)
        return body

    def by_text(self, text: str, *, node_name: str | None = None) -> bs4.Tag | None:
        nodes = self.html.find_all(string=re.compile(rf"\s*{text}\s*"))
        for node in nodes:
            if isinstance(node, bs4.NavigableString):
                node = node.parent

            if node_name:
                while node is not None:
                    if node.name == node_name:
                        return node
                    node = node.parent

        return None

    def by_label_text(self, text: str) -> bs4.Tag | None:
        label = self.by_text(text, node_name="label")
        assert label is not None
        assert label.attrs.get("for") is not None
        resp = self.html.find(id=label.attrs["for"])
        assert not isinstance(resp, bs4.NavigableString)
        return resp

    @property
    def form(self) -> WebForm:
        if self._form is None:
            form = self.html.form
            assert form is not None
            self._form = WebForm(self._client, self._origin, form)
        return self._form

    def by_node_name(
        self, node_name: str, *, attrs: dict[str, str] | None = None
    ) -> list[bs4.Tag]:
        return self.html.find_all(node_name, attrs or {})


CookieTypes = httpx._types.CookieTypes  # type: ignore
Cookies = httpx._models.Cookies  # type: ignore


class WebTestClient:
    def __init__(
        self,
        app: ASGIApp,
        cookies: CookieTypes | None = None,
    ) -> None:
        self.app = app
        self.testclient = TestClient(app, cookies=cookies or {})

    @property
    def cookies(self) -> Cookies:
        return self.testclient.cookies

    @cookies.setter
    def cookies(self, value: Cookies) -> None:
        self.testclient.cookies = value

    def get(self, url: str) -> WebResponse:
        resp = self.testclient.get(url, follow_redirects=False)
        if "set-cookie" in resp.headers:
            for name, cookie in resp.cookies.items():
                self.testclient.cookies.set(name, cookie)
        return WebResponse(
            self,
            url,
            resp,
        )

    def post(self, url: str, data: Mapping[str, Any]) -> WebResponse:
        return WebResponse(
            self,
            url,
            self.testclient.post(
                url,
                content=urlencode(data),
                headers={"Content-Type": "application/x-www-form-urlencoded"},
                follow_redirects=False,
            ),
        )
